from __future__ import unicode_literals
from django_cradmin.renderable import AbstractRenderable


class AbstractItemRenderer(AbstractRenderable):
    """
    Abstract base class for :class:`.ItemFrameRenderer`
    and :class:`.InnerValueRenderer`.

    You will normally add objects of a class that extends this class
    to :class:`.List`, but it is not required.
    """

    #: If this is specified, we will add an attribute with this name
    #: for the value as an attribute of the object.
    #:
    #: I.e.: if ``valuealias = "person"``, you will be able to use ``me.person`` in
    #: the template, and you will be able to use ``self.person`` in any methods you
    #: add or override in the class (just remember to call ``super()`` if you override
    #: ``__init__``).
    valuealias = None

    def __init__(self, value):
        """
        Parameters:
            value: The value to render.
        """
        self.value = value
        if self.valuealias:
            setattr(self, self.valuealias, self.value)


class ItemValueRenderer(AbstractItemRenderer):
    """
    The value renderer renders the value of each item in
    the :class:`.List`.
    """
    template_name = 'django_cradmin/viewhelpers/listbuilder/base/itemvalue.django.html'

    def get_css_classes(self):
        """
        Override this to set your own css classes.
        """
        return 'django-cradmin-listbuilder-itemvalue'


class ItemFrameRenderer(AbstractItemRenderer):
    """
    The outer item renderer provides a frame around the item.

    This is typically subclassed to:

    - Add visually highlighted frames for some items.
    - Add checkbox for each item
    - Make the item clickable (via an ``a`` tag or javascript).

    .. note::

        The :class:`.List` dors not require a :class:`.ItemFrameRenderer`,
        so lightweight lists without this extra wrapper is possible.
    """
    template_name = 'django_cradmin/viewhelpers/listbuilder/base/itemframe.django.html'

    def __init__(self, inneritem):
        super(ItemFrameRenderer, self).__init__(inneritem.value)
        self.inneritem = inneritem

    def get_css_classes(self):
        """
        Override this to set your own css classes.
        """
        return 'django-cradmin-listbuilder-itemframe'


class List(AbstractRenderable):
    """
    A builder for HTML lists (can be used for other lists as well).

    Each item in the list is a :class:`.AbstractItemRenderer` object, and
    it is typically:

    - A subclass of :class:`.ItemValueRenderer` for simple lists.
    - A subclass of :class:`.ItemFrameRenderer` for more complex lists.

    .. note:: Items can be any :class:`django_cradmin.renderable.AbstractRenderable`, but that does not
        work with the shortcut methods:

        - :meth:`.extend_with_values`
        - :meth:`.from_value_iterable`
    """
    template_name = 'django_cradmin/viewhelpers/listbuilder/base/list.django.html'

    def __init__(self, extra_css_classes=None):
        self.renderable_list = []
        self.extra_css_classes = extra_css_classes

    def iter_renderables(self):
        return iter(self.renderable_list)

    def has_items(self):
        return len(self.renderable_list) > 0

    def get_css_classes(self):
        """
        Override this to set your own css classes.
        """
        css_classes = 'django-cradmin-listbuilder-list'
        if self.extra_css_classes:
            css_classes += ' ' + self.extra_css_classes
        return css_classes

    def append(self, renderable):
        """
        Appends an item to the list.

        Parameters:
            renderable: Must implement the :class:`django_cradmin.renderable.AbstractRenderable` interface,
                and it is typically a subclass of :class:`.AbstractItemRenderer`.
        """
        self.renderable_list.append(renderable)

    def extend(self, renerable_iterable):
        """
        Just like :meth:`.append` except that it takes an iterable
        of renderables instead of a single renderable.
        """
        self.renderable_list.extend(renerable_iterable)

    def get_default_value_renderer_class(self):
        """
        The default ``value_renderer_class`` for :meth:`.extend_with_values`.
        """
        return ItemValueRenderer

    def get_default_frame_renderer_class(self):
        """
        The default ``frame_renderer_class`` for :meth:`.extend_with_values`.
        """
        return None

    def extend_with_values(self, value_iterable, value_renderer_class=None, frame_renderer_class=None):
        """
        Extends the list with an iterable of values.

        Parameters:
            value_iterable: An iterable of values to add to the list.
            value_renderer_class: A subclass of :class:`.ItemValueRenderer`.
                Defaults to :meth:`.get_default_value_renderer_class`.
            frame_renderer_class: A subclass of :class:`.ItemFrameRenderer`.

                If this is ``None``, the object added to the list will be an object of the
                ``value_renderer_class``.

                If this is specified, the object added to the list will be an object
                of the ``frame_renderer_class`` with an object of the ``value_renderer_class``
                as the ``inneritem`` attribute/parameter.

                Defaults to :meth:`.get_default_frame_renderer_class`.
        """
        value_renderer_class = value_renderer_class or self.get_default_value_renderer_class()
        frame_renderer_class = frame_renderer_class or self.get_default_frame_renderer_class()
        if frame_renderer_class:
            renderable_iterable = [frame_renderer_class(inneritem=value_renderer_class(value=value))
                                   for value in value_iterable]
        else:
            renderable_iterable = map(value_renderer_class, value_iterable)
        self.extend(renderable_iterable)

    @classmethod
    def from_value_iterable(cls, value_iterable,
                            value_renderer_class=None,
                            frame_renderer_class=None,
                            **listkwargs):
        """
        A shortcut for creating an object of this class with the given ``**listkwargs``
        as __init__ arguments, and then calling :meth:`.extend_with_values` with
        ``value_iterable``, ``value_renderer_class`` and ``frame_renderer_class`` as
        arguments.

        Returns:
            An object of this class.
        """
        listobject = cls(**listkwargs)
        listobject.extend_with_values(value_iterable, value_renderer_class,
                                      frame_renderer_class=frame_renderer_class)
        return listobject



from django.template import RequestContext
from django.template.loader import render_to_string


class AbstractRenderable(object):
    """
    An abstract class that implements an interface for
    rendering something.

    Everything is just helpers for the :meth:`.render` method,
    which renders a template with an object of this class as
    input.
    """
    template_name = None

    def get_template_name(self):
        """
        Get the template name for :meth:`.render`.

        Defaults to :obj:`~.AbstractRenderable.template_name`.

        Raises:
            NotImplementedError: If :obj:`~.AbstractRenderable.template_name` is
             not set.
        """
        if self.template_name:
            return self.template_name
        else:
            raise NotImplementedError('You must set template_name or override '
                                      'get_template_name().')

    def get_context_data(self):
        """
        Get context data for :meth:`.render`.

        Defaults to::

            {
                'me': self
            }
        """
        return {
            'me': self
        }

    def get_template_context_object(self, request=None):
        """
        Get the template context object returned by
        :meth:`.render`.

        See the docs for the ``request`` parameter for
        :meth:`.render` for more details.
        """
        context = self.get_context_data()
        if request:
            context = RequestContext(request, context)
        return context

    def render(self, request=None):
        """
        Render :obj:`.get_template_name` with
        the context returned by :meth:`.get_context_data`.

        Paramteters:
            request (HttpRequest): If this is provided, we create
                wrap the response from :meth:`.get_context_data`
                in a ``RequestContext``.

                You can override this behavior in
                :meth:`.get_template_context_object`.
        """
        return render_to_string(
            self.get_template_name(),
            self.get_template_context_object(request=request))



from django.template import RequestContext
from django.test import TestCase, RequestFactory
import htmls
from django_cradmin.viewhelpers import listbuilder


class MinimalRenderable(listbuilder.base.AbstractRenderable):
    template_name = 'cradmin_viewhelpers_testapp/listbuilder/minimal-renderable.django.html'


class TestAbstractRenderable(TestCase):
    def test_get_template_name_no_template_name(self):
        with self.assertRaises(NotImplementedError):
            listbuilder.base.AbstractRenderable().get_template_name()

    def test_get_template_name_has_template_name(self):
        MinimalRenderable().get_template_name()  # NotImplementedError not raised

    def test_get_context_data(self):
        renderable = listbuilder.base.AbstractRenderable()
        self.assertEqual({'me': renderable}, renderable.get_context_data())

    def test_get_template_context_object_without_request(self):
        renderable = listbuilder.base.AbstractRenderable()
        self.assertEqual({'me': renderable}, renderable.get_template_context_object())

    def test_get_template_context_object_with_request(self):
        renderable = listbuilder.base.AbstractRenderable()
        request = RequestFactory().get('')
        context_object = renderable.get_template_context_object(request)
        self.assertTrue(isinstance(context_object, RequestContext))
        self.assertEqual(renderable, context_object['me'])

    def test_render(self):
        self.assertEqual('Test', MinimalRenderable().render().strip())


class TestAbstractItemRenderer(TestCase):
    def test_without_valuealias(self):
        itemrenderer = listbuilder.base.AbstractItemRenderer(value='testvalue')
        self.assertEqual('testvalue', itemrenderer.value)

    def test_with_valuealias(self):
        class MyItemRenderer(listbuilder.base.AbstractItemRenderer):
            valuealias = 'myvalue'

        itemrenderer = MyItemRenderer(value='testvalue')
        self.assertEqual('testvalue', itemrenderer.value)
        self.assertEqual('testvalue', itemrenderer.myvalue)


class TestItemValueRenderer(TestCase):
    def test_render(self):
        rendered = listbuilder.base.ItemValueRenderer(value='testvalue').render()
        selector = htmls.S(rendered)
        self.assertEqual('testvalue',
                         selector.one('.django-cradmin-listbuilder-itemvalue').alltext_normalized)


class TestItemFrameRenderer(TestCase):
    def test_render(self):
        rendered = listbuilder.base.ItemFrameRenderer(
            inneritem=listbuilder.base.ItemValueRenderer(value='testvalue')).render()
        selector = htmls.S(rendered)
        self.assertTrue(selector.exists('.django-cradmin-listbuilder-itemframe'))
        self.assertEqual('testvalue',
                         selector.one('.django-cradmin-listbuilder-itemvalue').alltext_normalized)


class TestList(TestCase):
    def test_append(self):
        testlist = listbuilder.base.List()
        renderable = MinimalRenderable()
        testlist.append(renderable)
        self.assertEqual([renderable], testlist.renderable_list)

    def test_extend(self):
        testlist = listbuilder.base.List()
        renderable1 = MinimalRenderable()
        renderable2 = MinimalRenderable()
        testlist.extend([renderable1, renderable2])
        self.assertEqual([renderable1, renderable2], testlist.renderable_list)

    def test_extend_with_values_without_frame_renderer_class(self):
        testlist = listbuilder.base.List()
        testlist.extend_with_values(['testvalue1', 'testvalue2'],
                                    value_renderer_class=listbuilder.base.ItemValueRenderer)
        self.assertEqual('testvalue1', testlist.renderable_list[0].value)
        self.assertEqual('testvalue2', testlist.renderable_list[1].value)
        self.assertTrue(isinstance(testlist.renderable_list[0], listbuilder.base.ItemValueRenderer))
        self.assertTrue(isinstance(testlist.renderable_list[1], listbuilder.base.ItemValueRenderer))

    def test_extend_with_values_with_frame_renderer_class(self):
        testlist = listbuilder.base.List()
        testlist.extend_with_values(['testvalue1', 'testvalue2'],
                                    value_renderer_class=listbuilder.base.ItemValueRenderer,
                                    frame_renderer_class=listbuilder.base.ItemFrameRenderer)
        self.assertEqual('testvalue1', testlist.renderable_list[0].value)
        self.assertEqual('testvalue2', testlist.renderable_list[1].value)
        self.assertTrue(isinstance(testlist.renderable_list[0], listbuilder.base.ItemFrameRenderer))
        self.assertTrue(isinstance(testlist.renderable_list[1], listbuilder.base.ItemFrameRenderer))

    def test_from_value_iterable_without_frame_renderer_class(self):
        testlist = listbuilder.base.List.from_value_iterable(
            ['testvalue1', 'testvalue2'],
            value_renderer_class=listbuilder.base.ItemValueRenderer)
        self.assertEqual('testvalue1', testlist.renderable_list[0].value)
        self.assertEqual('testvalue2', testlist.renderable_list[1].value)
        self.assertTrue(isinstance(testlist.renderable_list[0], listbuilder.base.ItemValueRenderer))
        self.assertTrue(isinstance(testlist.renderable_list[1], listbuilder.base.ItemValueRenderer))

    def test_from_value_iterable_with_frame_renderer_class(self):
        testlist = listbuilder.base.List.from_value_iterable(
            ['testvalue1', 'testvalue2'],
            value_renderer_class=listbuilder.base.ItemValueRenderer,
            frame_renderer_class=listbuilder.base.ItemFrameRenderer)
        self.assertEqual('testvalue1', testlist.renderable_list[0].value)
        self.assertEqual('testvalue2', testlist.renderable_list[1].value)
        self.assertTrue(isinstance(testlist.renderable_list[0], listbuilder.base.ItemFrameRenderer))
        self.assertTrue(isinstance(testlist.renderable_list[1], listbuilder.base.ItemFrameRenderer))

    def test_render(self):
        testlist = listbuilder.base.List()
        testlist.append(listbuilder.base.ItemValueRenderer(value='testvalue1'))
        testlist.append(listbuilder.base.ItemValueRenderer(value='testvalue2'))
        selector = htmls.S(testlist.render())
        self.assertEqual(2, selector.count('li'))
        self.assertEqual('testvalue1', selector.list('li')[0].alltext_normalized)
        self.assertEqual('testvalue2', selector.list('li')[1].alltext_normalized)



from django_cradmin.viewhelpers.listbuilder import base


class RowList(base.List):
    def get_css_classes(self):
        css_classes = super(RowList, self).get_css_classes()
        css_classes += ' ' + 'django-cradmin-listbuilder-rowlist'
        return css_classes


class FloatGridList(base.List):

    def get_css_size_class(self):
        return 'django-cradmin-listbuilder-floatgridlist-md'

    def get_css_classes(self):
        css_classes = super(FloatGridList, self).get_css_classes()
        css_classes += ' django-cradmin-listbuilder-floatgridlist ' + self.get_css_size_class()
        return css_classes



# from django_cradmin.viewhelpers.listbuilder import base
#
#
# class Link(base.ItemValueRenderer):
#     """
#     Renders a value item as a link.
#     """
#     def get_url(self):
#         """
#         Get the URL the link should go to.
#
#         Defaults to ``value.get_absolute_url``, which means that this
#         works out of the box with any Django model object with
#         the ``get_absolute_url`` method defined.
#         """
#         return self.value.get_absolute_url()



from django_cradmin.viewhelpers.listbuilder import base


class Link(base.ItemFrameRenderer):
    """
    Renders a frame as a link.
    """
    template_name = 'django_cradmin/viewhelpers/listbuilder/itemframe/link.django.html'

    def get_css_classes(self):
        css_classes = super(Link, self).get_css_classes()
        return css_classes + ' django-cradmin-listbuilder-itemframe-link'

    def get_url(self):
        """
        Get the URL the link should go to.

        Defaults to ``value.get_absolute_url``, which means that this
        works out of the box with any Django model object with
        the ``get_absolute_url`` method defined.
        """
        return self.value.get_absolute_url()



from django.conf import settings
from django.template import defaultfilters
from django.views.generic import ListView
from django_cradmin.viewhelpers import listbuilder
from django.utils.translation import ugettext_lazy as _


class ViewMixin(object):
    """
    Listbuilder view mixin. Must be mixin in before any Django View subclass.

    This is typically used with a Django ListView or TemplateView.
    The mixin is not dependent on any specific backend, so it works
    no matter where you get your data from (database, mongodb, elasticsearch, ...).

    For a ready to use view that extends this to work with Django model objects,
    see :class:`.View`.

    The ViewMixin works much like :class:`.View`, but you must override/implement:

    - :meth:`.get_pagetitle`
    - :meth:`.get_listbuilder_list_data`
    - :meth:`.get_no_items_message`

    Examples:

        Minimal example with data from a list::

            class

    """
    template_name = 'django_cradmin/viewhelpers/listbuilderview.django.html'

    #: See :meth:`~ViewMixin.get_listbuilder_class`.
    listbuilder_class = listbuilder.list.RowList

    #: See :meth:`~ViewMixin.get_value_renderer_class`.
    value_renderer_class = listbuilder.base.ItemValueRenderer

    #: See :meth:`~ViewMixin.get_frame_renderer_class`.
    frame_renderer_class = None

    #: Set this to True to hide the page header. See :meth:`~.FormViewMixin.get_hide_page_header`.
    hide_page_header = False

    def get_pagetitle(self):
        """
        Get the page title (the title tag).

        Must be implemented in subclasses.
        """
        raise NotImplementedError()

    def get_pageheading(self):
        """
        Get the page heading.

        Defaults to :meth:`.get_pagetitle`.
        """
        return self.get_pagetitle()

    def get_hide_page_header(self):
        """
        Return ``True`` if we should hide the page header.

        You can override this, or set :obj:`.hide_page_header`, or hide the page header
        in all form views with the ``DJANGO_CRADMIN_HIDE_PAGEHEADER_IN_LISTVIEWS`` setting.
        """
        return self.hide_page_header or getattr(settings, 'DJANGO_CRADMIN_HIDE_PAGEHEADER_IN_LISTVIEWS', False)

    def get_listbuilder_class(self):
        """
        Get a subclass of :class:`django_cradmin.viewhelpers.listbuilder.base.List`.

        Defaults to :obj:`.ViewMixin.listbuilder_class`.
        """
        return self.listbuilder_class

    def get_listbuilder_list_kwargs(self):
        """
        Get kwargs for :meth:`.get_listbuilder_class`.
        """
        return {}

    def get_value_renderer_class(self):
        """
        Get a subclass of :class:`django_cradmin.viewhelpers.listbuilder.base.ItemValueRenderer`.

        Defaults to :obj:`.ViewMixin.value_renderer_class`.
        """
        return self.value_renderer_class

    def get_frame_renderer_class(self):
        """
        Get a subclass of :class:`django_cradmin.viewhelpers.listbuilder.base.ItemFrameRenderer`.

        Defaults to :obj:`.ViewMixin.frame_renderer_class`.
        """
        return self.frame_renderer_class

    def get_listbuilder_list_value_iterable(self, context):
        """
        Get the value_iterable for the listbuilder list.

        Must be overridden in subclasses.

        Parameters:
            context: The Django template context.
        """
        raise NotImplementedError()

    def get_listbuilder_list(self, context):
        """
        Get the listbuilder List object.

        You normally do not have to override this, but instead you should
        override:

        - :meth:`.get_listbuilder_list_value_iterable`
        - :meth:`.get_value_renderer_class`
        - :meth:`.get_frame_renderer_class`
        - :meth:`.get_listbuilder_list_kwargs`

        Parameters:
            context: The Django template context.
        """
        return self.get_listbuilder_class().from_value_iterable(
            value_iterable=self.get_listbuilder_list_value_iterable(context),
            value_renderer_class=self.get_value_renderer_class(),
            frame_renderer_class=self.get_frame_renderer_class(),
            **self.get_listbuilder_list_kwargs())

    def get_no_items_message(self):
        """
        Get the message to show when there are no items.

        Must be overridden in subclasses.
        """
        raise NotImplementedError()

    def add_listview_context_data(self, context):
        context['listbuilder_list'] = self.get_listbuilder_list(context)
        context['pagetitle'] = self.get_pagetitle()
        context['hide_pageheader'] = self.get_hide_page_header()
        context['pageheading'] = self.get_pageheading()
        context['no_items_message'] = self.get_no_items_message()

    def get_context_data(self, **kwargs):
        context = super(ViewMixin, self).get_context_data(**kwargs)
        self.add_listview_context_data(context)
        return context


class View(ViewMixin, ListView):
    """
    View using the :doc:`viewhelpers_listbuilder`.

    Examples:

        Minimal::

            class MyView(listbuilderview.View):
                def get_queryset(self):
                    return MyModel.objects.all()

    """

    #: The model class to list objects for. You do not have to specify
    #: this, but if you do not specify this or :meth:`~.ObjectTableView.get_model_class`,
    #: you have to override :meth:`~.ObjectTableView.get_pagetitle` and
    #: :meth:`~.ObjectTableView.get_no_items_message`.
    model = None

    def get_model_class(self):
        """
        Get the model class to list objects for.

        Defaults to :obj:`.model`. See :obj:`.model` for more info.
        """
        return self.model

    def get_pagetitle(self):
        """
        Get the page title (the title tag).

        Defaults to the ``verbose_name_plural`` of the :obj:`.model`
        with the first letter capitalized.
        """
        return defaultfilters.capfirst(self.get_model_class()._meta.verbose_name_plural)

    def get_listbuilder_list_value_iterable(self, context):
        return context['object_list']

    def get_queryset_for_role(self, role):
        """
        Get a queryset with all objects of :obj:`.model`  that
        the current role can access.
        """
        raise NotImplementedError()

    def get_queryset(self):
        """
        DO NOT override this. Override :meth:`.get_queryset_for_role`
        instead.
        """
        queryset = self.get_queryset_for_role(self.request.cradmin_role)
        return queryset

    def get_no_items_message(self):
        """
        Get the message to show when there are no items.
        """
        return _('No %(modelname_plural)s') % {
            'modelname_plural': self.get_model_class()._meta.verbose_name_plural.lower(),
        }
